""" Script for generating statistics from sensitive workload data.

It would produce aggregated data per group (privatized). And finally per day 
and hour statistics.

The idea of this script is to create a release of aggregated statistics 
which does not allow for unprivatizing of the gid (which for me is SLURM's
account).

This script will produce:

    (invariant, that is unanonymized) 
    1) per week submit time % of jobs
    2) per hour submit time % of jobs
    
    (anonymized)
    3) % of jobs submitted per group 
    
    per group
    
    4) runtime histogram
    5) allocated cpus histogram

Copyright (C) 2024
Manuel G. Marciani
BSC-CNS Earth Sciences Department
"""

# library to handle differential privacy
#import opendp.prelude as dp
# in-house library for handling workload format
import pyswf

def invariant_statistics(swftrace: pyswf.SWFtrace) -> None:
    """ This function computes all the statistics that are not meant to be 
    altered by anonimization, hence called invariants.

    It will produce plots for the weekly and hourly submit time of the machine.
    Those will be saved to files.

    swftrace: the SWFtrace object with the workload data 
    """
    # for plots
    import matplotlib.pyplot as plt
    # for timedelta calculations for each one of the timestamps 
    from datetime import timedelta

    # invariants (not privatized released): submit, start, end, uid, gid (account)
    # we are computing (in timestamp) each of the instants, for then joining per day of the week 
    swftrace.job_list["submit_ts"] = swftrace.job_list.submit_time.apply(lambda instant: timedelta(seconds=instant)) + swftrace.start_time
    swftrace.job_list["start_ts"] = swftrace.job_list.start_time.apply(lambda instant: timedelta(seconds=instant)) + swftrace.start_time
    swftrace.job_list["end_ts"] = swftrace.job_list.end_time.apply(lambda instant: timedelta(seconds=instant)) + swftrace.start_time

    # submit info
    # weekday will be 0, Monday, 1, Tuesday, 2 Wednesday, ...
    swftrace.job_list["submit_day_of_the_week"] = swftrace.job_list["submit_ts"].apply(lambda date: date.weekday())
    swftrace.job_list["submit_hour_of_the_day"] = swftrace.job_list["submit_ts"].apply(lambda date: date.hour)

    # aggregated counts
    no_jobs_submit_agg_per_day_week = swftrace.job_list.groupby("submit_day_of_the_week").count()
    no_jobs_submit_agg_per_hour_of_the_day = swftrace.job_list.groupby("submit_hour_of_the_day").count()

    days_in_week = ['M','Tu','W','Th','F','Sat','Sun']
    hours_in_day = [f'{hour}' for hour in range(24)]

    plt.plot(days_in_week, no_jobs_submit_agg_per_day_week/len(workload.job_list))
    plt.title("Percentage of submitted jobs per day of the week") 
    plt.ylabel("Percentage of jobs")
    plt.savefig("no_jobs_submit_agg_per_day_week.png", dpi=300)
    plt.clf()
   
    plt.plot(hours_in_day, no_jobs_submit_agg_per_hour_of_the_day/len(workload.job_list))
    plt.title("Percentage of submitted jobs per hour of the day") 
    plt.ylabel("Percentage of jobs")   
    plt.savefig("no_jobs_submit_agg_per_hour_of_the_day.png", dpi=300)
    plt.clf()

def anonymized_statistics(swftrace: pyswf.SWFtrace, epsilon=1.0) -> None:
    """ Create the anonymized statistics and output them both in graphs and the raw data

    swftrace: input standard workflow format object
    epsilon: anonymization factor
    """
    
    # assuming symetric distance in the input domain
    max_influence = 1 # this is the maximum changes that the privatized entity might cause on the (end) distribution
    budget = (1., 1e-8) # output distance assuming it is laplace :) 
    
    #private: per group (slurm account) runtime histogram (in percentage). bounded to 0 and max(runtime)
    #private: per group (slurm account) allocated_cpus (in percentage) histogram bounded to 0 and max_procs
    #private: per group (slurm account) % of jobs from account (sum
    gids = swftrace.job_list.gid.unique()

    # these are the columns of each of the dataframes with the implicit column 
    # as well "df_id"
    col_names = ["df_id"] + swftrace.job_list.columns()

    # this is pretty ugly, but works. We will convert the dataframe to csv
    # this is just a deterministic transformation.
    per_gid_num_jobs_histogram = (
        dp.t.make_split_dateframe(separator=",", col_names=col_names) >>
        dp.t.make_select_column(key="gid",TOA=str) >>
        dp.t.then_cast(TOA=int) >>
        dp.t.then_count_by_categories(categories=gids)
    )
    
    noisy_per_gid_num_jobs_histogram = dp.binary_search_chain(
        lambda s: histogram >> dp.m.then_laplace(scale=s),
        d_in=max_influence, d_out=budget[0]
    )

    #alloc_cpus_preprocessor = (
    #    dp.t.make_select_column(key='alloc_cpus', TOA=str)
    #) 
    #alloc_cpus_cast_str_int = (
    #    alloc_cpus_preprocessor.output_space >>
    #    dp.t.then_cast(TOA=int) 
    #    dp.t.then_clamp(0, swftrace.max_procs) # limit from 0 to max_procs (maximum number of cpus) the allocated cpus
    #)
    #df_alloc_cpus = workload.job_list.alloc_cpus

if __name__ == "__main__":
    # load data
    workload = pyswf.SWFtrace.from_swf("/home/mgimenez/Documents/local/wrapper paper/workloads/CEA-Curie/original/CEA-Curie-2011-2.1-cln.swf")
    
    invariant_statistics(workload)
   
